import wandb
import glob
import os

import numpy as np

from torch.nn.utils.rnn import pad_sequence

from ..utils import create_exp_dir
from ..metrics import MetricMeter, AverageMeter, AUCROC


class Logger:
    def __init__(self, config, wandb=False):
        exp_name = config["exp_name"]
        self.wandb = wandb

        if wandb:
            wandb.init(project="MIMIC Benchmark", name=exp_name)
            wandb.config.update(config)
            wandb.run.log_code("*.py")

        self.experiment_path = f"./exp/{exp_name}"

        if not os.path.exists("./exp"):
            os.mkdir("./exp")

        create_exp_dir(self.experiment_path, scripts_to_save=glob.glob("*.py"))
        np.save(os.path.join(self.experiment_path, "config"), config)

        self.metrics = {
            "Loss": AverageMeter(),
            "AUC-ROC Micro": MetricMeter(AUCROC("micro")),
            "AUC-ROC Macro": MetricMeter(AUCROC("macro")),
        }

    def update(self, outputs, labels, loss):
        batch_size = outputs.size(0)

        label_tmp = labels.cpu().numpy()
        outputs = outputs.cpu().detach().numpy()

        self.metrics["Loss"].update(loss.item(), batch_size)
        self.metrics["AUC-ROC Micro"].update(label_tmp, outputs)
        self.metrics["AUC-ROC Macro"].update(label_tmp, outputs)



import torch
import torch.nn as nn
from torch import optim
from torch.utils.data import DataLoader
from torch.nn.utils.rnn import pad_sequence

from torchmimic.data import DecompensationDataset
from torchmimic.utils import pad_colalte

from torchmimic.loggers import DecompensationLogger


class DecompensationBenchmark:
    def __init__(
        self,
        model,
        train_batch_size=8,
        test_batch_size=256,
        data="/data/datasets/mimic3-benchmarks/data/decompensation",
        learning_rate=0.001,
        weight_decay=0,
        report_freq=200,
        exp_name="Test",
        device="cpu",
        sample_size=None,
        workers=5,
        wandb=False,
    ):

        self.test_batch_size = test_batch_size
        self.train_batch_size = train_batch_size
        self.learning_rate = learning_rate
        self.weight_decay = weight_decay
        self.model = model
        self.device = device
        self.report_freq = report_freq

        config = {}
        config.update(model.get_config())
        config.update(self.get_config())

        self.logger = DecompensationLogger(exp_name + "_decomp", config, wandb)

        torch.cuda.set_device(self.device)

        train_dataset = DecompensationDataset(
            data,
            train=True,
            n_samples=sample_size,
        )

        test_dataset = DecompensationDataset(
            data,
            train=False,
            n_samples=sample_size,
        )

        kwargs = (
            {"num_workers": workers, "pin_memory": True} if self.device else {}
        )

        self.train_loader = DataLoader(
            train_dataset,
            batch_size=train_batch_size,
            shuffle=True,
            collate_fn=pad_colalte,
            **kwargs,
        )
        self.test_loader = DataLoader(
            test_dataset,
            batch_size=test_batch_size,
            shuffle=False,
            collate_fn=pad_colalte,
            **kwargs,
        )

        self.model = self.model.to(self.device)

        self.optimizer = optim.Adam(
            model.parameters(),
            lr=learning_rate,
            weight_decay=weight_decay,
            betas=(0.9, 0.98),
        )

        self.crit = nn.BCELoss()

    def fit(self, epochs):

        for epoch in range(epochs):
            self.model.train()
            self.logger.reset()
            for batch_idx, (data, label, lens, mask) in enumerate(
                self.train_loader
            ):
                data = data.to(self.device)
                label = label.to(self.device)

                output = self.model((data, lens))
                loss = self.crit(output[:, 0], label)
                loss.backward()
                self.optimizer.step()
                self.optimizer.zero_grad(set_to_none=True)

                self.logger.update(output, label, loss)

                if (batch_idx + 1) % self.report_freq == 0:
                    print(
                        f"Train: epoch: {epoch+1}, loss = {self.logger.get_loss()}"
                    )

            self.logger.print_metrics(epoch, split="Train")

            self.model.eval()
            self.logger.reset()
            with torch.no_grad():
                for batch_idx, (data, label, lens, mask) in enumerate(
                    self.test_loader
                ):
                    data = data.to(self.device)
                    label = label.to(self.device)

                    output = self.model((data, lens))
                    loss = self.crit(output[:, 0], label)

                    self.logger.update(output, label, loss)

                    if (batch_idx + 1) % self.report_freq == 0:
                        print(
                            f"Eval: epoch: {epoch+1}, loss = {self.logger.get_loss()}"
                        )

                self.logger.print_metrics(epoch, split="Eval")

    def get_config(self):
        return {
            "test_batch_size": self.test_batch_size,
            "train_batch_size": self.train_batch_size,
            "learning_rate": self.learning_rate,
            "weight_decay": self.weight_decay,
        }



import glob
import os

from abc import ABC, abstractmethod

import numpy as np
import torch
import wandb

from torchmimic.metrics import AverageMeter, MetricMeter
from torchmimic.utils import create_exp_dir


class BaseLogger(ABC):
    """
    Base Logger class. Used for logging, printing, and saving information about the run. Contains built-in wandb support.

    :param config: A dictionary of the run configuration
    :type config: dict
    :param log_wandb: If true, wandb will be used to log metrics and configuration
    :type log_wandb: bool
    """

    def __init__(self, exp_name, config, log_wandb=False):
        """
        Initialize BaseLogger

        :param config: A dictionary of the run configuration
        :type config: dict
        :param log_wandb: If true, wandb will be used to log metrics and configuration
        :type log_wandb: bool
        """
        self.log_wandb = log_wandb

        if self.log_wandb:
            wandb.init(project="MIMIC Benchmark", name=exp_name)
            wandb.config.update(config)
            wandb.run.log_code("*.py")

        self.experiment_path = f"./exp/{exp_name}"

        if not os.path.exists("./exp"):
            os.mkdir("./exp")

        create_exp_dir(self.experiment_path, scripts_to_save=glob.glob("*.py"))
        np.save(os.path.join(self.experiment_path, "config"), config)

        self.metrics = {
            "Loss": AverageMeter(),
        }

    def __del__(self):
        """
        Destructor for BaseLogger. Finishes wandb if log_wandb is true
        """
        if self.log_wandb:
            wandb.finish()

    @abstractmethod
    def update(self, outputs, labels, loss):
        """
        Abstract class for updating metrics
        """
        pass

    def reset(self):
        """
        Resets metrics
        """
        for item in self.metrics.values():
            item.reset()

    def get_loss(self):
        """
        Returns average loss

        :return: Average Loss
        :rtype: float
        """
        return self.metrics["Loss"].avg

    def print_metrics(self, epoch, split="Train"):
        """
        Prints and logs metrics. If log_wandb is True, wandb run will be updated

        :param epoch: The current epoch
        :type epoch: int
        :param split: The split of the data. "Train" or "Eval"
        :type split: str
        """

        assert split in ("Train", "Eval")

        result_str = split + ": "

        if self.log_wandb:
            wandb.log({"Epoch": epoch + 1}, commit=False)

        result_str += f" Epoch {epoch+1}"
        for name, meter in self.metrics.items():
            if isinstance(meter, MetricMeter):
                result = meter.score()
            elif isinstance(meter, AverageMeter):
                result = meter.avg

            if self.log_wandb:
                wandb.log({split + " " + name: result}, commit=False)
            result_str += f", {name}={result}"

        print(result_str)
        if split == "Eval" and self.log_wandb:
            wandb.log({})

    def save(self, model):
        """
        Saves the provides models to the experiment path
        """
        torch.save(
            model.state_dict(),
            os.path.join(self.experiment_path, "weights.pt"),
        )



from torchmimic.metrics import MetricMeter, AUCROC
from .base_logger import BaseLogger


class PhenotypingLogger(BaseLogger):
    """
    Phenotyping Logger class. Used for logging, printing, and saving information about the run. Logs loss, AUC-ROC macro, and AUC-ROC micro. Contains built-in wandb support.

    :param config: A dictionary of the run configuration
    :type config: dict
    :param log_wandb: If true, wandb will be used to log metrics and configuration
    :type log_wandb: bool
    """

    def __init__(self, exp_name, config, log_wandb=False):
        """
        Initialize PhenotypingLogger

        :param config: A dictionary of the run configuration
        :type config: dict
        :param log_wandb: If true, wandb will be used to log metrics and configuration
        :type log_wandb: bool
        """

        super().__init__(exp_name, config, log_wandb=log_wandb)

        self.metrics.update(
            {
                "AUC-ROC Micro": MetricMeter(AUCROC("micro")),
                "AUC-ROC Macro": MetricMeter(AUCROC("macro")),
            }
        )

    def update(self, outputs, labels, loss):
        """
        Update loss, AUC-ROC macro, and AUC-ROC micro

        :param outputs: Predicted labels
        :type outputs: torch.Tensor
        :param labels: True labels
        :type labels: torch.Tensor
        :param loss: Loss from the training iteration.
        :type loss: float
        """

        batch_size = outputs.size(0)

        label_tmp = labels.cpu().numpy()
        outputs = outputs.cpu().detach().numpy()

        self.metrics["Loss"].update(loss.item(), batch_size)
        self.metrics["AUC-ROC Micro"].update(label_tmp, outputs)
        self.metrics["AUC-ROC Macro"].update(label_tmp, outputs)



import os
import random
import numpy as np


class Reader:
    def __init__(self, dataset_dir, listfile=None):
        self._dataset_dir = dataset_dir
        self._current_index = 0
        if listfile is None:
            listfile_path = os.path.join(dataset_dir, "listfile.csv")
        else:
            listfile_path = listfile
        with open(listfile_path, "r") as lfile:
            self._data = lfile.readlines()
        self._listfile_header = self._data[0]
        self._data = self._data[1:]

    def get_number_of_examples(self):
        return len(self._data)

    def random_shuffle(self, seed=None):
        if seed is not None:
            random.seed(seed)
        random.shuffle(self._data)

    def read_example(self, index):
        raise NotImplementedError()

    def read_next(self):
        to_read_index = self._current_index
        self._current_index += 1
        if self._current_index == self.get_number_of_examples():
            self._current_index = 0
        return self.read_example(to_read_index)


class DecompensationReader(Reader):
    def __init__(self, dataset_dir, listfile=None):
        """Reader for decompensation prediction task.
        :param dataset_dir: Directory where timeseries files are stored.
        :param listfile:    Path to a listfile. If this parameter is left `None` then
                            `dataset_dir/listfile.csv` will be used.
        """
        Reader.__init__(self, dataset_dir, listfile)
        self._data = [line.split(",") for line in self._data]
        self._data = [(x, float(t), int(y)) for (x, t, y) in self._data]

    def _read_timeseries(self, ts_filename, time_bound):
        ret = []
        with open(os.path.join(self._dataset_dir, ts_filename), "r") as tsfile:
            header = tsfile.readline().strip().split(",")
            assert header[0] == "Hours"
            for line in tsfile:
                mas = line.strip().split(",")
                t = float(mas[0])
                if t > time_bound + 1e-6:
                    break
                ret.append(np.array(mas))
        return (np.stack(ret), header)

    def read_example(self, index):
        """Read the example with given index.
        :param index: Index of the line of the listfile to read (counting starts from 0).
        :return: Directory with the following keys:
            X : np.array
                2D array containing all events. Each row corresponds to a moment.
                First column is the time and other columns correspond to different
                variables.
            t : float
                Length of the data in hours. Note, in general, it is not equal to the
                timestamp of last event.
            y : int (0 or 1)
                Mortality within next 24 hours.
            header : array of strings
                Names of the columns. The ordering of the columns is always the same.
            name: Name of the sample.
        """
        if index < 0 or index >= len(self._data):
            raise ValueError(
                "Index must be from 0 (inclusive) to number of examples (exclusive)."
            )

        name = self._data[index][0]
        t = self._data[index][1]
        y = self._data[index][2]
        (X, header) = self._read_timeseries(name, t)

        return {"X": X, "t": t, "y": y, "header": header, "name": name}


class InHospitalMortalityReader(Reader):
    def __init__(self, dataset_dir, listfile=None, period_length=48.0):
        """Reader for in-hospital moratality prediction task.
        :param dataset_dir:   Directory where timeseries files are stored.
        :param listfile:      Path to a listfile. If this parameter is left `None` then
                              `dataset_dir/listfile.csv` will be used.
        :param period_length: Length of the period (in hours) from which the prediction is done.
        """
        Reader.__init__(self, dataset_dir, listfile)
        self._data = [line.split(",") for line in self._data]
        self._data = [(x, int(y)) for (x, y) in self._data]
        self._period_length = period_length

    def _read_timeseries(self, ts_filename):
        ret = []
        with open(os.path.join(self._dataset_dir, ts_filename), "r") as tsfile:
            header = tsfile.readline().strip().split(",")
            assert header[0] == "Hours"
            for line in tsfile:
                mas = line.strip().split(",")
                ret.append(np.array(mas))
        return (np.stack(ret), header)

    def read_example(self, index):
        """Reads the example with given index.
        :param index: Index of the line of the listfile to read (counting starts from 0).
        :return: Dictionary with the following keys:
            X : np.array
                2D array containing all events. Each row corresponds to a moment.
                First column is the time and other columns correspond to different
                variables.
            t : float
                Length of the data in hours. Note, in general, it is not equal to the
                timestamp of last event.
            y : int (0 or 1)
                In-hospital mortality.
            header : array of strings
                Names of the columns. The ordering of the columns is always the same.
            name: Name of the sample.
        """
        if index < 0 or index >= len(self._data):
            raise ValueError(
                "Index must be from 0 (inclusive) to number of lines (exclusive)."
            )

        name = self._data[index][0]
        t = self._period_length
        y = self._data[index][1]
        (X, header) = self._read_timeseries(name)

        return {"X": X, "t": t, "y": y, "header": header, "name": name}


class LengthOfStayReader(Reader):
    def __init__(self, dataset_dir, listfile=None):
        """Reader for length of stay prediction task.
        :param dataset_dir: Directory where timeseries files are stored.
        :param listfile:    Path to a listfile. If this parameter is left `None` then
                            `dataset_dir/listfile.csv` will be used.
        """
        Reader.__init__(self, dataset_dir, listfile)
        self._data = [line.split(",") for line in self._data]
        self._data = [(x, float(t), float(y)) for (x, t, y) in self._data]

    def _read_timeseries(self, ts_filename, time_bound):
        ret = []
        with open(os.path.join(self._dataset_dir, ts_filename), "r") as tsfile:
            header = tsfile.readline().strip().split(",")
            assert header[0] == "Hours"
            for line in tsfile:
                mas = line.strip().split(",")
                t = float(mas[0])
                if t > time_bound + 1e-6:
                    break
                ret.append(np.array(mas))
        return (np.stack(ret), header)

    def read_example(self, index):
        """Reads the example with given index.
        :param index: Index of the line of the listfile to read (counting starts from 0).
        :return: Dictionary with the following keys:
            X : np.array
                2D array containing all events. Each row corresponds to a moment.
                First column is the time and other columns correspond to different
                variables.
            t : float
                Length of the data in hours. Note, in general, it is not equal to the
                timestamp of last event.
            y : float
                Remaining time in ICU.
            header : array of strings
                Names of the columns. The ordering of the columns is always the same.
            name: Name of the sample.
        """
        if index < 0 or index >= len(self._data):
            raise ValueError(
                "Index must be from 0 (inclusive) to number of lines (exclusive)."
            )

        name = self._data[index][0]
        t = self._data[index][1]
        y = self._data[index][2]
        (X, header) = self._read_timeseries(name, t)

        return {"X": X, "t": t, "y": y, "header": header, "name": name}


class PhenotypingReader(Reader):
    def __init__(self, dataset_dir, listfile=None):
        """Reader for phenotype classification task.
        :param dataset_dir: Directory where timeseries files are stored.
        :param listfile:    Path to a listfile. If this parameter is left `None` then
                            `dataset_dir/listfile.csv` will be used.
        """
        Reader.__init__(self, dataset_dir, listfile)
        self._data = [line.split(",") for line in self._data]
        self._data = [
            (mas[0], float(mas[1]), list(map(int, mas[2:])))
            for mas in self._data
        ]

    def _read_timeseries(self, ts_filename):
        ret = []
        with open(os.path.join(self._dataset_dir, ts_filename), "r") as tsfile:
            header = tsfile.readline().strip().split(",")
            assert header[0] == "Hours"
            for line in tsfile:
                mas = line.strip().split(",")
                ret.append(np.array(mas))
        return (np.stack(ret), header)

    def read_example(self, index):
        """Reads the example with given index.
        :param index: Index of the line of the listfile to read (counting starts from 0).
        :return: Dictionary with the following keys:
            X : np.array
                2D array containing all events. Each row corresponds to a moment.
                First column is the time and other columns correspond to different
                variables.
            t : float
                Length of the data in hours. Note, in general, it is not equal to the
                timestamp of last event.
            y : array of ints
                Phenotype labels.
            header : array of strings
                Names of the columns. The ordering of the columns is always the same.
            name: Name of the sample.
        """
        if index < 0 or index >= len(self._data):
            raise ValueError(
                "Index must be from 0 (inclusive) to number of lines (exclusive)."
            )

        name = self._data[index][0]
        t = self._data[index][1]
        y = self._data[index][2]
        (X, header) = self._read_timeseries(name)

        return {"X": X, "t": t, "y": y, "header": header, "name": name}

    def __len__(self):
        return self.get_number_of_examples()

    def __getitem__(self, index):
        return self.read_example(index)


class MultitaskReader(Reader):
    def __init__(self, dataset_dir, listfile=None):
        """Reader for multitask learning.
        :param dataset_dir: Directory where timeseries files are stored.
        :param listfile:    Path to a listfile. If this parameter is left `None` then
                            `dataset_dir/listfile.csv` will be used.
        """
        Reader.__init__(self, dataset_dir, listfile)
        self._data = [line.split(",") for line in self._data]

        def process_ihm(x):
            return list(map(int, x.split(";")))

        def process_los(x):
            x = x.split(";")
            if x[0] == "":
                return ([], [])
            return (
                list(map(int, x[: len(x) // 2])),
                list(map(float, x[len(x) // 2 :])),
            )

        def process_ph(x):
            return list(map(int, x.split(";")))

        def process_decomp(x):
            x = x.split(";")
            if x[0] == "":
                return ([], [])
            return (
                list(map(int, x[: len(x) // 2])),
                list(map(int, x[len(x) // 2 :])),
            )

        self._data = [
            (
                fname,
                float(t),
                process_ihm(ihm),
                process_los(los),
                process_ph(pheno),
                process_decomp(decomp),
            )
            for fname, t, ihm, los, pheno, decomp in self._data
        ]

    def _read_timeseries(self, ts_filename):
        ret = []
        with open(os.path.join(self._dataset_dir, ts_filename), "r") as tsfile:
            header = tsfile.readline().strip().split(",")
            assert header[0] == "Hours"
            for line in tsfile:
                mas = line.strip().split(",")
                ret.append(np.array(mas))
        return (np.stack(ret), header)

    def read_example(self, index):
        """Reads the example with given index.
        :param index: Index of the line of the listfile to read (counting starts from 0).
        :return: Return dictionary with the following keys:
            X : np.array
                2D array containing all events. Each row corresponds to a moment.
                First column is the time and other columns correspond to different
                variables.
            t : float
                Length of the data in hours. Note, in general, it is not equal to the
                timestamp of last event.
            ihm : array
                Array of 3 integers: [pos, mask, label].
            los : array
                Array of 2 arrays: [masks, labels].
            pheno : array
                Array of 25 binary integers (phenotype labels).
            decomp : array
                Array of 2 arrays: [masks, labels].
            header : array of strings
                Names of the columns. The ordering of the columns is always the same.
            name: Name of the sample.
        """
        if index < 0 or index >= len(self._data):
            raise ValueError(
                "Index must be from 0 (inclusive) to number of lines (exclusive)."
            )

        name = self._data[index][0]
        (X, header) = self._read_timeseries(name)

        return {
            "X": X,
            "t": self._data[index][1],
            "ihm": self._data[index][2],
            "los": self._data[index][3],
            "pheno": self._data[index][4],
            "decomp": self._data[index][5],
            "header": header,
            "name": name,
        }



import platform
import pickle
import json
import os
import numpy as np


class Discretizer:
    def __init__(
        self,
        timestep=0.8,
        store_masks=True,
        impute_strategy="zero",
        start_time="zero",
        config_path=os.path.join(
            os.path.dirname(__file__),
            "discretizers",
            "discretizer_config.json",
        ),
    ):

        with open(config_path) as f:
            config = json.load(f)
            self._id_to_channel = config["id_to_channel"]
            self._channel_to_id = dict(
                zip(self._id_to_channel, range(len(self._id_to_channel)))
            )
            self._is_categorical_channel = config["is_categorical_channel"]
            self._possible_values = config["possible_values"]
            self._normal_values = config["normal_values"]

        self._header = ["Hours"] + self._id_to_channel
        self._timestep = timestep
        self._store_masks = store_masks
        self._start_time = start_time
        self._impute_strategy = impute_strategy

        # for statistics
        self._done_count = 0
        self._empty_bins_sum = 0
        self._unused_data_sum = 0

    def transform(self, X, header=None, end=None):
        if header is None:
            header = self._header
        assert header[0] == "Hours"
        eps = 1e-6

        N_channels = len(self._id_to_channel)
        ts = [float(row[0]) for row in X]
        for i in range(len(ts) - 1):
            assert ts[i] < ts[i + 1] + eps

        if self._start_time == "relative":
            first_time = ts[0]
        elif self._start_time == "zero":
            first_time = 0
        else:
            raise ValueError("start_time is invalid")

        if end is None:
            max_hours = max(ts) - first_time
        else:
            max_hours = end - first_time

        N_bins = int(max_hours / self._timestep + 1.0 - eps)

        cur_len = 0
        begin_pos = [0 for i in range(N_channels)]
        end_pos = [0 for i in range(N_channels)]
        for i in range(N_channels):
            channel = self._id_to_channel[i]
            begin_pos[i] = cur_len
            if self._is_categorical_channel[channel]:
                end_pos[i] = begin_pos[i] + len(self._possible_values[channel])
            else:
                end_pos[i] = begin_pos[i] + 1
            cur_len = end_pos[i]

        data = np.zeros(shape=(N_bins, cur_len), dtype=float)
        mask = np.zeros(shape=(N_bins, N_channels), dtype=int)
        original_value = [
            ["" for j in range(N_channels)] for i in range(N_bins)
        ]
        total_data = 0
        unused_data = 0

        def write(data, bin_id, channel, value, begin_pos):
            channel_id = self._channel_to_id[channel]
            if self._is_categorical_channel[channel]:
                category_id = self._possible_values[channel].index(value)
                N_values = len(self._possible_values[channel])
                one_hot = np.zeros((N_values,))
                one_hot[category_id] = 1
                for pos in range(N_values):
                    data[bin_id, begin_pos[channel_id] + pos] = one_hot[pos]
            else:
                data[bin_id, begin_pos[channel_id]] = float(value)

        for row in X:
            t = float(row[0]) - first_time
            if t > max_hours + eps:
                continue
            bin_id = int(t / self._timestep - eps)
            assert 0 <= bin_id < N_bins

            for j in range(1, len(row)):
                if row[j] == "":
                    continue
                channel = header[j]
                channel_id = self._channel_to_id[channel]

                total_data += 1
                if mask[bin_id][channel_id] == 1:
                    unused_data += 1
                mask[bin_id][channel_id] = 1

                write(data, bin_id, channel, row[j], begin_pos)
                original_value[bin_id][channel_id] = row[j]

        # impute missing values

        if self._impute_strategy not in [
            "zero",
            "normal_value",
            "previous",
            "next",
        ]:
            raise ValueError("impute strategy is invalid")

        if self._impute_strategy in ["normal_value", "previous"]:
            prev_values = [[] for i in range(len(self._id_to_channel))]
            for bin_id in range(N_bins):
                for channel in self._id_to_channel:
                    channel_id = self._channel_to_id[channel]
                    if mask[bin_id][channel_id] == 1:
                        prev_values[channel_id].append(
                            original_value[bin_id][channel_id]
                        )
                        continue
                    if self._impute_strategy == "normal_value":
                        imputed_value = self._normal_values[channel]
                    if self._impute_strategy == "previous":
                        if len(prev_values[channel_id]) == 0:
                            imputed_value = self._normal_values[channel]
                        else:
                            imputed_value = prev_values[channel_id][-1]
                    write(data, bin_id, channel, imputed_value, begin_pos)

        if self._impute_strategy == "next":
            prev_values = [[] for i in range(len(self._id_to_channel))]
            for bin_id in range(N_bins - 1, -1, -1):
                for channel in self._id_to_channel:
                    channel_id = self._channel_to_id[channel]
                    if mask[bin_id][channel_id] == 1:
                        prev_values[channel_id].append(
                            original_value[bin_id][channel_id]
                        )
                        continue
                    if len(prev_values[channel_id]) == 0:
                        imputed_value = self._normal_values[channel]
                    else:
                        imputed_value = prev_values[channel_id][-1]
                    write(data, bin_id, channel, imputed_value, begin_pos)

        empty_bins = np.sum(
            [1 - min(1, np.sum(mask[i, :])) for i in range(N_bins)]
        )
        self._done_count += 1
        self._empty_bins_sum += empty_bins / (N_bins + eps)
        self._unused_data_sum += unused_data / (total_data + eps)

        if self._store_masks:
            data = np.hstack([data, mask.astype(np.float32)])

        # create new header
        new_header = []
        for channel in self._id_to_channel:
            if self._is_categorical_channel[channel]:
                values = self._possible_values[channel]
                for value in values:
                    new_header.append(channel + "->" + value)
            else:
                new_header.append(channel)

        if self._store_masks:
            for i in range(len(self._id_to_channel)):
                channel = self._id_to_channel[i]
                new_header.append("mask->" + channel)

        new_header = ",".join(new_header)

        return (data, new_header)

    def print_statistics(self):
        print("statistics of discretizer:")
        print("\tconverted {} examples".format(self._done_count))
        print(
            "\taverage unused data = {:.2f} percent".format(
                100.0 * self._unused_data_sum / self._done_count
            )
        )
        print(
            "\taverage empty  bins = {:.2f} percent".format(
                100.0 * self._empty_bins_sum / self._done_count
            )
        )


class Normalizer:
    def __init__(self, fields=None):
        self._means = None
        self._stds = None
        self._fields = None
        if fields is not None:
            self._fields = [col for col in fields]

        self._sum_x = None
        self._sum_sq_x = None
        self._count = 0

    def _feed_data(self, x):
        x = np.array(x)
        self._count += x.shape[0]
        if self._sum_x is None:
            self._sum_x = np.sum(x, axis=0)
            self._sum_sq_x = np.sum(x**2, axis=0)
        else:
            self._sum_x += np.sum(x, axis=0)
            self._sum_sq_x += np.sum(x**2, axis=0)

    def _save_params(self, save_file_path):
        eps = 1e-7
        with open(save_file_path, "wb") as save_file:
            N = self._count
            self._means = 1.0 / N * self._sum_x
            self._stds = np.sqrt(
                1.0
                / (N - 1)
                * (
                    self._sum_sq_x
                    - 2.0 * self._sum_x * self._means
                    + N * self._means**2
                )
            )
            self._stds[self._stds < eps] = eps
            pickle.dump(
                obj={"means": self._means, "stds": self._stds},
                file=save_file,
                protocol=2,
            )

    def load_params(self, load_file_path):
        with open(load_file_path, "rb") as load_file:
            if platform.python_version()[0] == "2":
                dct = pickle.load(load_file)
            else:
                dct = pickle.load(load_file, encoding="latin1")
            self._means = dct["means"]
            self._stds = dct["stds"]

    def transform(self, X):
        if self._fields is None:
            fields = range(X.shape[1])
        else:
            fields = self._fields
        ret = 1.0 * X
        for col in fields:
            ret[:, col] = (X[:, col] - self._means[col]) / self._stds[col]
        return ret



from .decompensation.dataset import DecompensationDataset
from .ihm.dataset import IHMDataset
from .los.dataset import LOSDataset
from .phenotyping.dataset import PhenotypingDataset

# from .multitask.dataset import MultitaaskDataset



import numpy as np

from sklearn.metrics import (
    f1_score,
    roc_auc_score,
    balanced_accuracy_score,
    mean_absolute_error,
    precision_recall_curve,
    auc,
    cohen_kappa_score,
)


def kappa(true, pred):
    """
    Returns the Cohen's Kappa for the provided true and predicted values

    :param true: true values
    :type true: np.array
    :param pred: predicted values
    :type pred: np.array
    :return: Cohen's Kappa score
    :rtype: int
    """
    pred = np.argmax(pred, axis=1)
    true = true[:, 0]
    return cohen_kappa_score(true, pred, weights="linear")


def accuracy(true, pred):
    """
    Returns the accuracy for the provided true and predicted values

    :param true: true values
    :type true: np.array
    :param pred: predicted values
    :type pred: np.array
    :return: accuracy score
    :rtype: int
    """
    return (sum(true == pred) / len(pred)) * 100


def f1(true, pred):
    """
    Returns the F1-score for the provided true and predicted values

    :param true: true values
    :type true: np.array
    :param pred: predicted values
    :type pred: np.array
    :return: F1-score
    :rtype: int
    """
    return f1_score(true, pred)


def balanced_accuracy(true, pred):
    """
    Returns the Balanced Accuracy for the provided true and predicted values

    :param true: true values
    :type true: np.array
    :param pred: predicted values
    :type pred: np.array
    :return: Balanced Accuracy score
    :rtype: int
    """
    return balanced_accuracy_score(true, pred)


def mae(true, pred):
    """
    Returns the Mean Absolute Error/Deviation for the provided
    true and predicted values

    :param true: true values
    :type true: np.array
    :param pred: predicted values
    :type pred: np.array
    :return: MAE/MAD score
    :rtype: int
    """
    one_hot = np.zeros((true.size, true.max() + 1))
    for i in np.arange(true.size):
        one_hot[np.arange(true.size), true[i]] = 1
    return mean_absolute_error(one_hot, pred)


def aucpr(true, pred):
    """
    Returns the AUC-PR for the provided true and predicted values

    :param true: true values
    :type true: np.array
    :param pred: predicted values
    :type pred: np.array
    :return: AUC-PR score
    :rtype: int
    """
    (precisions, recalls, _) = precision_recall_curve(true, pred)
    return auc(recalls, precisions)


class AUCROC:
    """
    AUCROC scoring class
    """

    def __init__(self, average=None):
        """
        Initialization for AUCROC class

        :param average: type of average used for multiclass.
        :type average: str
        """
        self.average = average

    def __call__(self, true, pred):
        """
        Returns the AUC-ROC for the provided true and predicted values

        :param true: true values
        :type true: np.array
        :param pred: predicted values
        :type pred: np.array
        :return: AUC-ROC score
        :rtype: int
        """
        return roc_auc_score(
            true, pred, multi_class="ovr", average=self.average
        )


class AverageMeter:
    """
    Class used to collect values and return a running average
    """

    def __init__(self):
        """
        Initializae the AverageMeter class
        """
        self.avg = 0
        self.sum = 0
        self.cnt = 0

    def reset(self):
        """
        Resets private members
        """
        self.avg = 0
        self.sum = 0
        self.cnt = 0

    def update(self, val, _n=1):
        """
        Updates class members

        :param val: value used to update running average
        :type val: float
        :param _n: sample size used to calculate value
        :type _n: int
        """
        self.sum += val * _n
        self.cnt += _n
        self.avg = self.sum / self.cnt


class MetricMeter:
    """
    Class used to collect values and evaluate them using a scoring function
    """

    def __init__(self, score_fn):
        """
        Initializae the MetricMeter Class

        :param score_fn: scoring function
        :type score_fn: function
        """
        super().__init__()
        self.reset()
        self.score_fn = score_fn

    def reset(self):
        """
        Resets private members
        """
        self.true = []
        self.pred = []

    def update(self, true, pred):
        """
        Updates list of true and predicted values

        :param true: true labels
        :type true: np.array
        :param pred: predicted labels
        :type pred: np.array
        """
        self.pred.append(pred)
        self.true.append(true)

    def score(self):
        """
        Scores true and predicted values
        :returns: the output of the score function given the predicted and true labels
        :rtype: int
        """
        self.true = np.concatenate(self.true, axis=0)
        self.pred = np.concatenate(self.pred, axis=0)
        return self.score_fn(self.true, self.pred)



from __future__ import absolute_import
from __future__ import print_function

import torch

import numpy as np

import random
import os

from ..utils import get_bin_custom, read_chunk


class BatchGen(object):
    def __init__(
        self,
        reader,
        discretizer,
        normalizer,
        batch_size,
        small_part,
        shuffle,
        partition,
        return_mask=False,
    ):
        self.discretizer = discretizer
        self.normalizer = normalizer
        self.ihm_pos = int(48.0 / 1 - 1e-6)
        self.partition = partition
        self.batch_size = batch_size
        self.shuffle = shuffle

        self._load_data(reader, discretizer, normalizer, small_part)

        self.steps = len(self.data)

    def _load_data(self, reader, discretizer, normalizer, small_part=False):
        N = reader.get_number_of_examples()
        if small_part:
            N = 1000

        ret = read_chunk(reader, N)
        Xs = ret["X"]
        ts = ret["t"]
        ihms = ret["ihm"]
        loss = ret["los"]
        phenos = ret["pheno"]
        decomps = ret["decomp"]

        self.data = dict()
        self.data["pheno_ts"] = ts
        self.data["names"] = ret["name"]
        self.data["decomp_ts"] = []
        self.data["los_ts"] = []

        for i in range(N):
            self.data["decomp_ts"].append(
                [pos for pos, m in enumerate(decomps[i][0]) if m == 1]
            )
            self.data["los_ts"].append(
                [pos for pos, m in enumerate(loss[i][0]) if m == 1]
            )
            (
                Xs[i],
                ihms[i],
                decomps[i],
                loss[i],
                phenos[i],
            ) = self._preprocess_single(
                Xs[i], ts[i], ihms[i], decomps[i], loss[i], phenos[i]
            )

        self.data["X"] = Xs
        self.data["ihm_M"] = [x[0] for x in ihms]
        self.data["ihm_y"] = [x[1] for x in ihms]
        self.data["decomp_M"] = [x[0] for x in decomps]
        self.data["decomp_y"] = [x[1] for x in decomps]
        self.data["los_M"] = [x[0] for x in loss]
        self.data["los_y"] = [x[1] for x in loss]
        self.data["pheno_y"] = phenos

        self.seq_lens = [len(Xs[0]) for x in ihms]

    def _preprocess_single(self, X, max_time, ihm, decomp, los, pheno):
        timestep = self.discretizer._timestep
        eps = 1e-6

        def get_bin(t):
            return int(t / timestep - eps)

        n_steps = get_bin(max_time) + 1

        # X
        X = self.discretizer.transform(X, end=max_time)[0]
        if self.normalizer is not None:
            X = self.normalizer.transform(X)
        assert len(X) == n_steps

        # ihm
        # NOTE: when mask is 0, we set y to be 0. This is important
        #       because in the multitask networks when ihm_M = 0 we set
        #       our prediction thus the loss will be 0.
        if np.equal(ihm[1], 0):
            ihm[2] = 0
        ihm = (np.int32(ihm[1]), np.int32(ihm[2]))  # mask, label

        # decomp
        decomp_M = [0] * n_steps
        decomp_y = [0] * n_steps
        for i in range(len(decomp[0])):
            pos = get_bin(i)
            decomp_M[pos] = decomp[0][i]
            decomp_y[pos] = decomp[1][i]
        decomp = (
            np.array(decomp_M, dtype=np.int32),
            np.array(decomp_y, dtype=np.int32),
        )

        # los
        los_M = [0] * n_steps
        los_y = [0] * n_steps
        for i in range(len(los[0])):
            pos = get_bin(i)
            los_M[pos] = los[0][i]
            los_y[pos] = los[1][i]
        los = (
            np.array(los_M, dtype=np.int32),
            np.array(los_y, dtype=np.float32),
        )

        # pheno
        pheno = np.array(pheno, dtype=np.int32)

        return (X, ihm, decomp, los, pheno)

    def get_max_seq_length(self):
        return max(self.seq_lens)

    def __getitem__(self, idx):
        outputs = []

        # X
        X = self.data["X"][idx]
        T = X.shape[1]

        ihm_M = np.array(self.data["ihm_M"][idx])
        ihm_M = np.expand_dims(ihm_M, axis=-1)  # (B, 1)
        ihm_y = np.array(self.data["ihm_y"][idx])
        ihm_y = torch.FloatTensor()  # (B, 1)
        outputs.append(ihm_y)

        # decomp
        decomp_M = self.data["decomp_M"][idx]
        decomp_y = self.data["decomp_y"][idx]
        decomp_y = torch.FloatTensor(
            np.expand_dims(decomp_y, axis=-1)
        )  # (B, T, 1)
        outputs.append(decomp_y)

        # los
        los_M = self.data["los_M"][idx]
        los_y = self.data["los_y"][idx]

        los_y = np.array([get_bin_custom(ly, 10) for ly in los_y], dtype=int)
        los_y = torch.FloatTensor(los_y)
        outputs.append(los_y)

        # pheno
        pheno_y = torch.FloatTensor(self.data["pheno_y"][idx])
        outputs.append(pheno_y)

        inputs = [X, ihm_M, decomp_M, los_M]

        return inputs, outputs, self.seq_lens[idx]

    def __len__(self):
        return self.steps



from .decompensation.decompensation_benchmark import DecompensationBenchmark
from .ihm.ihm_benchmark import IHMBenchmark
from .los.los_benchmark import LOSBenchmark
from .phenotyping.phenotyping_benchmark import PhenotypingBenchmark



import os
import shutil
import torch

import numpy as np
from torch.nn.utils.rnn import pad_sequence


def pad_colalte(batch):
    xx, yy, lens, mask = zip(*batch)
    x = pad_sequence(xx, batch_first=True)
    y = torch.stack(yy, dim=0)

    return x, y, lens, mask


def create_exp_dir(path, scripts_to_save=None):
    if not os.path.exists(path):
        os.mkdir(path)
    print(f"Experiment dir: {path}")

    if scripts_to_save is not None:
        if not os.path.exists(os.path.join(path, "scripts")):
            os.mkdir(os.path.join(path, "scripts"))
        for script in scripts_to_save:
            dst_file = os.path.join(path, "scripts", os.path.basename(script))
            shutil.copyfile(script, dst_file)



import torch


def read_chunk(reader, chunk_size):
    data = {}
    for _ in range(chunk_size):
        ret = reader.read_next()
        for k, v in ret.items():
            if k not in data:
                data[k] = []
            data[k].append(v)
    data["header"] = data["header"][0]
    return data


class CustomBins:
    inf = 1e18
    bins = [
        (-inf, 1),
        (1, 2),
        (2, 3),
        (3, 4),
        (4, 5),
        (5, 6),
        (6, 7),
        (7, 8),
        (8, 14),
        (14, +inf),
    ]
    nbins = len(bins)
    means = [
        11.450379,
        35.070846,
        59.206531,
        83.382723,
        107.487817,
        131.579534,
        155.643957,
        179.660558,
        254.306624,
        585.325890,
    ]


def get_bin_custom(x, nbins, one_hot=False):
    for i in range(nbins):
        a = CustomBins.bins[i][0] * 24.0
        b = CustomBins.bins[i][1] * 24.0
        if a <= x < b:
            if one_hot:
                ret = torch.zeros((CustomBins.nbins,))
                ret[i] = 1
                return int(ret)
            return torch.Tensor([i]).long()
    return None



from __future__ import absolute_import
from __future__ import print_function

import random
import os
import torch

import numpy as np

from torchmimic.data.preprocessing import Discretizer, Normalizer
from torchmimic.data.readers import PhenotypingReader
from torchmimic.data.utils import read_chunk
from torchmimic.data.base_dataset import BaseDataset


class PhenotypingDataset(BaseDataset):
    """
    Phenotyping dataset that can be directly used by PyTorch dataloaders. This class preprocessing the data the same way as "Multitask learning and benchmarking with clinical time series data": https://github.com/YerevaNN/mimic3-benchmarks

    :param root: directory where data is located
    :type root: str
    :param train: if true, the training split of the data will be used. Otherwise, the validation dataset will be used
    :type train: bool
    :param n_samples: number of samples to use. If None, all the data is used
    :type steps: int
    """

    def __init__(
        self,
        root,
        train=True,
        transform=None,
        n_samples=None,
    ):
        """
        Initialize PhenotypingDataset

        :param root: directory where data is located
        :type root: str
        :param train: if true, the training split of the data will be used. Otherwise, the validation dataset will be used
        :type train: bool
        :param n_samples: number of samples to use. If None, all the data is used
        :type steps: int
        """
        super().__init__(transform=transform)

        listfile = "train_listfile.csv" if train else "val_listfile.csv"

        self._read_data(root, listfile)
        self._load_data(n_samples)

        self.n_samples = len(self.data)

    def _read_data(self, root, listfile):
        self.reader = PhenotypingReader(
            dataset_dir=os.path.join(root, "train"),
            listfile=os.path.join(root, listfile),
        )

        self.discretizer = Discretizer(
            timestep=1.0,
            store_masks=True,
            impute_strategy="previous",
            start_time="zero",
        )

        discretizer_header = self.discretizer.transform(
            self.reader.read_example(0)["X"]
        )[1].split(",")
        cont_channels = [
            i for (i, x) in enumerate(discretizer_header) if x.find("->") == -1
        ]

        self.normalizer = Normalizer(fields=cont_channels)
        normalizer_state = "../normalizers/ph_ts1.0.input_str:previous.start_time:zero.normalizer"
        normalizer_state = os.path.join(
            os.path.dirname(__file__), normalizer_state
        )
        self.normalizer.load_params(normalizer_state)



import torch
import torch.nn as nn

from torch.nn.utils.rnn import pack_padded_sequence, pad_packed_sequence


class StandardLSTM(nn.Module):
    def __init__(
        self,
        n_classes,
        hidden_dim=128,
        num_layers=1,
        dropout_rate=0,
        bidirectional=False,
    ):
        super().__init__()
        self.num_layers = num_layers
        self.dropout_rate = dropout_rate
        self.bidirectional = bidirectional

        if bidirectional:
            self.hidden_dim = hidden_dim // 2

        if num_layers == 1:
            dropout_rate = 0

        self.lstm_layer = nn.LSTM(
            76,
            self.hidden_dim,
            num_layers=num_layers,
            dropout=dropout_rate,
            bidirectional=bidirectional,
            batch_first=True,
        )

        linear_input = self.hidden_dim
        if bidirectional:
            linear_input *= 2

        self.final_layer = nn.Sequential(
            nn.Dropout(dropout_rate),
            nn.Linear(linear_input, n_classes),
            nn.Sigmoid(),
        )

    def forward(self, data):
        seq = data[0]
        lens = data[1]
        packed = pack_padded_sequence(
            seq, lens, batch_first=True, enforce_sorted=False
        )

        h_dim = 2 if self.bidirectional else 1

        z, (ht, ct) = self.lstm_layer(packed)

        seq_unpacked, lens_unpacked = pad_packed_sequence(z, batch_first=True)

        output = self.final_layer(
            torch.vstack(
                [
                    seq_unpacked[i, int(l) - 1]
                    for i, l in enumerate(lens_unpacked)
                ]
            )
        )

        return output

    def get_config(self):
        return {
            "hidden_dim": self.hidden_dim,
            "num_layers": self.num_layers,
            "dropout_rate": self.dropout_rate,
            "bidirectional": self.bidirectional,
        }



from __future__ import absolute_import
from __future__ import print_function

from abc import ABC, abstractmethod

import torch
import numpy

from torchmimic.data.utils import read_chunk


class BaseDataset(ABC):
    def __init__(self, transform):
        self.transform = transform
        self.reader = None
        self.discretizer = None
        self.normalizer = None

    @abstractmethod
    def _read_data(self, root, listfile):
        pass

    def _load_data(self, sample_size):
        N = self.reader.get_number_of_examples()

        if sample_size is None:
            sample_size = N

        ret = read_chunk(self.reader, sample_size)

        data = ret["X"]
        ts = ret["t"]
        ys = ret["y"]
        names = ret["name"]

        data_tmp = []
        self.mask = []

        for X, t in zip(data, ts):
            d = self.discretizer.transform(X, end=t)[0]
            data_tmp.append(d)
            self.mask.append(self.expand_mask(d[:, 59:]))

        # data = [self.discretizer.transform(X, end=t)[0] for (X, t) in zip(data, ts)]
        if self.normalizer is not None:
            self.data = [
                torch.Tensor(self.normalizer.transform(X)) for X in data_tmp
            ]
        ys = torch.FloatTensor(ys)
        self.labels = ys
        self.ts = ts
        self.names = names

    def __getitem__(self, idx):
        x = self.data[idx]
        sl = len(x)
        y = self.labels[idx]
        m = self.mask[idx]

        if self.transform:
            x = self.transform(x)

        return x, y, sl, m

    def __len__(self):
        return self.n_samples

    def expand_mask(self, mask):
        expanded_mask = torch.ones((mask.shape[0], 59))

        for i, pv in enumerate(self.discretizer._possible_values.values()):
            n_values = len(pv) if not pv == [] else 1
            for p in range(n_values):
                expanded_mask[:, p + i] = torch.from_numpy(mask[:, i])

        return expanded_mask



import os
import re
import setuptools

try:
    # for pip >= 10
    from pip._internal.req import parse_requirements
except ImportError:
    # for pip <= 9.0.3
    from pip.req import parse_requirements

# for simplicity we actually store the version in the __version__ attribute in the source
here = os.path.realpath(os.path.dirname(__file__))
with open(os.path.join(here, "torchmimic", "__init__.py")) as f:
    meta_match = re.search(
        r"^__version__ = ['\"]([^'\"]*)['\"]", f.read(), re.M
    )
    if meta_match:
        version = meta_match.group(1)
    else:
        raise RuntimeError("Unable to find __version__ string.")

required = [
    "torch",
    "scikit-learn",
    "numpy",
    "wandb",
]


def read(fname="README.md"):
    with open(
        os.path.join(os.path.dirname(__file__), fname), encoding="utf-8"
    ) as cfile:
        return cfile.read()


setuptools.setup(
    name="torchmimic",
    version=version,
    author="Ryan King",
    author_email="kingrc15@tamu.edu",
    description="MIMIC Benchmark in PyTorch.",
    license="MIT",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    url="https://github.com/kingrc15/torchmimic",
    packages=setuptools.find_packages(),
    package_data={
        "torchmimic": [
            "data/discretizers/discretizer_config.json",
            "data/normalizers/*.normalizer",
        ]
    },
    include_package_data=True,
    install_requires=required,
    extras_require={
        "local": ["pytest", "black", "setuptools", "sphinx"],
    },
    python_requires="~=3.6",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
    ],
)
